/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <sys/stat.h>

#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#error No gmp.h!
#endif /* HAVE_GMP_H */

#ifdef __USE_MCRYPT
#ifndef _MCRYPT_H
#define _MCRYPT_H
#include <mcrypt.h>
#endif
#endif

#include "thread.h"
#include "sha1.h"
#include "tiger.h"
#include "aes.h"
#include "arcfour.h"
#include "serpent.h"
#include "twofish.h"

#ifndef _CRYPT_H
#define _CRYPT_H


#define RIJNDAEL_128	0x01
#define TWOFISH_128	0x02
#define SERPENT_128	0x03
#define ARCFOUR_STREAM	0x04

#define ARCFOUR_JUMP	256


#define RIJNDAEL_128_BLOCK_SIZE	16
#define RIJNDAEL_128_KEY_SIZE	16
#define RIJNDAEL_128_IV_SIZE	16
#define TWOFISH_128_BLOCK_SIZE	16
#define TWOFISH_128_KEY_SIZE	16
#define TWOFISH_128_IV_SIZE	16
#define SERPENT_128_BLOCK_SIZE	16
#define SERPENT_128_KEY_SIZE	16
#define SERPENT_128_IV_SIZE	16

#define CIPHER_NAME(type) \
	((type) == RIJNDAEL_128)? "rijndael" : \
		((type) == TWOFISH_128)? "twofish" : "serpent"

#define DEFAULT_RANDOM		"/dev/urandom"


typedef struct cipher {
	union {
		#ifdef __USE_MCRYPT
		MCRYPT			mcrypt_td;
		#endif
		struct aes_ctx		aes;
		struct arcfour_ctx	arcfour;
		struct serpent_ctx	serpent;
		struct twofish_ctx	twofish;
	} __ctx;
	unsigned char __xor_block[16];	/* all of our ciphers operate in */
					/* 128bit blocks */
	void *__ctx_ptr;
	void (*__encrypt)(void *,unsigned int,unsigned char *,const unsigned char *);
	void (*__decrypt)(void *,unsigned int,unsigned char *,const unsigned char *);
	int type;
} cipher_ctx;


#ifdef __USE_MCRYPT
static pthread_once_t	__cipher_init_once;
static object_t		*__cipher_init_mutex;
void __cipher_init(void);
#endif


cipher_ctx *cipher_create(int,void *,void *);
void cipher_free(cipher_ctx *);
void cipher_init(cipher_ctx *,int,void *,void *);
void cipher_destroy(cipher_ctx *);
void cipher_encrypt(cipher_ctx *,void *,int);
void cipher_decrypt(cipher_ctx *,void *,int);
#define cipher_get_type(c)		(*c).type
#define cipher_get_key_size(t)		16
#define cipher_get_block_size(t)	16


#define MEMXOR(d,s,l)  do { \
	size_t i; \
	for (i = 0; i < (size_t)(l); i++) { \
		((unsigned char *)(d))[i] ^= ((unsigned char *)(s))[i]; \
	} \
} while (0)

#define SHA1_160	0x05
#define TIGER_192	0x06
#define TIGER_160	0x07
#define TIGER_128	0x08

#define SHA1_160_BLOCK_SIZE	64
#define SHA1_160_DIGEST_SIZE	20
#define TIGER_192_BLOCK_SIZE	64
#define TIGER_192_DIGEST_SIZE	24
#define TIGER_160_BLOCK_SIZE	64
#define TIGER_160_DIGEST_SIZE	20
#define TIGER_128_BLOCK_SIZE	64
#define TIGER_128_DIGEST_SIZE	16

#define HASH_BLOCK_SIZE_MAX 	64
#define HASH_DIGEST_SIZE_MAX	24

#define HASH_NAME(type) \
	((type) == SHA1_160)? "sha1" : "tiger"

#define FOLD_SHA1_160(d)	MEMXOR((d),((d)+10),10)
#define FOLD_TIGER_128(d)	MEMXOR((d),((d)+8),8)
#define FOLD_TIGER_160(d)	MEMXOR((d),((d)+10),10)
#define FOLD_TIGER_192(d)	MEMXOR((d),((d)+12),12)

#define FOLD_HASH(a,b)		switch((a)) { \
	case SHA1_160:	FOLD_SHA1_160((b)); break; \
	case TIGER_128:	FOLD_TIGER_128((b)); break; \
	case TIGER_160:	FOLD_TIGER_160((b)); break; \
	case TIGER_192:	FOLD_TIGER_192((b)); break; \
	default:	ERROR("cannot fold; bad hash!"); abort(); \
}


#define IPAD	0x36
#define OPAD	0x5C

#define HASH(type,digest,plaintext,size) do { \
	hash_ctx h; \
	hash_init((type),&h); \
	hash_update(&h,(plaintext),(size)); \
	hash_final(&h,(digest)); \
} while (0)

#define HMAC(type,digest,key,key_size,plaintext,plaintext_size) do { \
	hash_ctx h; \
	hmac_init((type),&h,(key),(key_size)); \
	hash_update(&h,(plaintext),(plaintext_size)); \
	hmac_final(&h,(digest)); \
} while (0)

typedef struct hash_ctx {
	int type;
	unsigned char ctx[MAX(sizeof(TIGER_CTX),sizeof(SHA_CTX))];
	int block_size;
	int digest_size;

	/* for HMAC */
	unsigned char key[MAX(SHA1_160_BLOCK_SIZE,TIGER_192_BLOCK_SIZE)];
	unsigned char ipad[MAX(SHA1_160_BLOCK_SIZE,TIGER_192_BLOCK_SIZE)];
	unsigned char opad[MAX(SHA1_160_BLOCK_SIZE,TIGER_192_BLOCK_SIZE)];
	int key_size;
} hash_ctx;

void hash_init(int,hash_ctx *);
void hash_update(hash_ctx *,void *,int);
void hash_final(hash_ctx *,void *);

void hmac_init(int,hash_ctx *,void *,int);
void hmac_update(hash_ctx *,void *,int);
void hmac_final(hash_ctx *,void *);

int hash_get_digest_size(int);
int hash_get_block_size(int);


/*
 * psuedo-randomness generation routines
 *
 */
#define PRAND_JUMP(r,b) do { \
	unsigned char buf[64]; ssize_t i = (ssize_t)(b); \
	while (i > 0) { \
		prand_get((r),buf,MIN(i,(ssize_t)sizeof(buf))); \
		i -= (ssize_t)sizeof(buf); \
	} \
} while (0)

#define PRAND_DEFAULT	ARCFOUR_STREAM;
#define PRAND_EXPAND(primitive,out,len,...)	do { \
	prand_ctx ctx; \
	prand_init(&ctx,primitive,__VA_ARGS__); \
	prand_get(&ctx,out,len); \
	prand_destroy(&ctx); \
} while(0)

struct __prand_hash {
	size_t key_len;		/* key length */
	size_t seed_len;	/* original seed length */
	size_t pos;		/* bytes read from pool; position */
	size_t digest_size;
	size_t pool_size;
	unsigned char *key;
	unsigned char *seed;	/* malloc()'d 2x's seed_len for concatenation;
				 * in normal operation seed will look like:
				 * previous digest + original seed
				 */
	unsigned char pool[MAX(SHA1_160_DIGEST_SIZE,TIGER_192_DIGEST_SIZE)*2];
};


typedef struct prand_ctx {
	int __primitive;
	union {
		struct __prand_hash hash;
		struct arcfour_ctx arcfour;
	} __state;
} prand_ctx;

void prand_init(prand_ctx *,int,...);
void prand_get(prand_ctx *,void *,size_t);
void prand_xor(prand_ctx *,void *,size_t);
void prand_destroy(prand_ctx *);


/*
 * random 
 */
struct {
	object_t object;
	int fd;
	struct stat st;
	char *path;
	prand_ctx prng;
} __rand_source;
void rand_init(void);
void rand_get(void *,size_t);


/*
 * Diffie-Hellman Key-exchange routines
 *
 */
#define DH_MAX_PRIME_BIT_SIZE   2048
#define DH_MAX_PRIME_BYTE_SIZE  (DH_MAX_PRIME_BIT_SIZE/8)
#define DH_MAX_SECRET_BYTE_SIZE DH_MAX_PRIME_BYTE_SIZE
#define DH_SAFE		0x00
#define DH_Y_UNSAFE	0x01
#define DH_X_UNSAFE	0x02
#define DH_Z_UNSAFE	0x04

typedef struct dh_ctx {
	int p_size_bits;
	int p_size_bytes;

	mpz_t g; /* generator	*/
	mpz_t p; /* prime	*/
	mpz_t x; /* private key	*/
	mpz_t y; /* public key	*/
	mpz_t z; /* secret	*/
} dh_ctx;

void dh_init(dh_ctx *);
void dh_generate_parameters(dh_ctx *,int,unsigned int);
void dh_set_prime(dh_ctx *,mpz_t);
void dh_set_generator(dh_ctx *,unsigned int);
void dh_generate_key(dh_ctx *);
void dh_compute_key(dh_ctx *,mpz_t);
int dh_check(dh_ctx *);
void dh_destroy(dh_ctx *);


/*
 * Key-exchange prime generation routines
 *
 */
#ifndef DEFAULT_PRIME
#define DEFAULT_PRIME PACKAGE ## ".ke_prime"
#endif
#define KE_PRIME_SIZE			128 	/* in bytes	*/
#define KE_PRIME_BIT_SIZE		(KE_PRIME_SIZE * 8)
#define KE_PUBKEY_SIZE			128	/* in bytes	*/
#define KE_PUBKEY_BIT_SIZE		(KE_PUBKEY_SIZE * 8)
#define KE_GENERATOR			3	/* default g	*/
#define KE_GENERATE_PRIME_INTERVAL	(10*60)	/* in seconds	*/
#define KE_SEED_SIZE			16	/* in bytes	*/
#define KDF_HASH	TIGER_192
#define KDF_SEED_SIZE	(KE_SEED_SIZE*2)


object_t __ke_mutex;
mpz_t __ke_prime;

void ke_init(void);
void ke_generate_prime(void);
void ke_get_prime(mpz_t);

void ke_load_prime(void);
void ke_save_prime(void);


#endif /* _CRYPT_H */
