/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#ifndef _LINK_H
#define _LINK_H 1

#include "common.h"
#include "packet.h"
#include "queue.h"
#include "crypt.h"

/*
 * Metric Review
 * ---------------------
 * second = 1
 * milli  = 1/1000
 * micro  = 1/1000000
 * nano   = 1/1000000000
 *
 */
#define LINK_PACKETS_PER_SECOND	10
#define LINK_BYTES_PER_SECOND	LINK_PACKETS_PER_SECOND * PACKET_SIZE
#define LINK_MICRO_DELAY	(1000000/LINK_PACKETS_PER_SECOND)
#define LINK_MICROSLEEP_COST	1
#define LINK_NANO_DELAY		(LINK_MICRO_DELAY * 1000)
#define LINK_NANOSLEEP_COST	(LINK_MICROSLEEP_COST * 1000)


#define LINK_MAX_CONNECTIONS	100
#define LINK_MAX_NULL		20
#define PACKET_CACHE_MAX	255
#define CHANNEL_MAX_CONNECTIONS	5
#define LINK_PACKET_IN_THRESHOLD(l)	((link->initiator)? 17 : 9)
#define LINK_PACKET_OUT_THRESHOLD(l)	((link->initiator)? 9 : 17)


#define LINKWOULDBLOCK	-1
#define LINKSHUTDOWN	-2
#define LINKFAILED	-3

#define CHANNELWOULDBLOCK	-1
#define CHANNELSHUTDOWN		-2
#define CHANNELFAILED		-3


#define MAGIC_HEADER_SIZE	8
#define MAGIC_HEADER_STATIC	{'A','n','o','n','N','e','t',PROTOCOL_MAJOR}
//#define CHECK_MAGIC_HEADER(h)	(memcmp((h),{'A','n','o','n','N','e','t',PROTOCOL_MAJOR},8) == 0)
#define CHECK_MAGIC_HEADER(h)	(memcmp((h),"AnonNet\0",8) == 0)
#define SET_MAGIC_HEADER(h)	do { memcpy((h),"AnonNet\0",8); } while (0)


#define SHUFFLE_LINK_LIST(r) do { /* prand context */ \
	int i; \
	unsigned char buf[LINK_MAX_CONNECTIONS]; \
	prand_get((r),buf,link_list.length); \
	for (i = 0; i < link_list.length; i++) { \
		SWAP(link_ctx *,link_list.list[0],link_list.list[((buf[i])%link_list.length)]); \
	} \
} while (0)


#define PACKET_CLEAN(p) do { \
	bzero((p)->data,PACKET_SIZE); \
	(p)->pos	= (p)->data; \
	(p)->encrypted	= false; \
	(p)->isnull	= false; \
} while (0)


struct packet_t;
typedef struct packet_t packet_t;

struct packet_t {
	char data[PACKET_SIZE];
	char *pos;
	long int link_id; /* link owner */
	bool encrypted;
	bool isnull;
	
	SLIST_ENTRY(packet_t) sle;
	SIMPLEQ_ENTRY(packet_t) sqe;
};


struct link_ctx;
typedef struct link_ctx link_ctx;

struct channel_ctx;
typedef struct channel_ctx channel_ctx;


struct link_ctx {
	object_t object;
	long int link_id;

	packet_t packet_alloc[PACKET_CACHE_MAX];
	int packet_alloc_len;

	SLIST_ENTRY(link_ctx) sle;

	int sd;
	SLIST_HEAD(,packet_t) cache;
	int cache_len;
	SIMPLEQ_HEAD(,packet_t) in;
	int in_len;
	size_t in_rcvd;
	size_t in_fail;
	size_t in_null;
	SIMPLEQ_HEAD(,packet_t) out;	
	int out_len;
	size_t out_sent;
	size_t out_fail;
	size_t out_part;
	size_t out_null;

	dh_ctx ke_dh;
	prand_ctx kdf;
	prand_ctx arc4_in;
	prand_ctx arc4_out;

	int hopid;
	bool iscaller;
	bool initiated;
	bool failed;
	bool closed;
};


struct channel_ctx {
	object_t object;
	long int channel_id;

	link_ctx *link;

	prand_ctx kdf;
	dh_ctx ke_dh;
	int hopid;
	cipher_ctx cipher_in;
	cipher_ctx cipher_out;
	unsigned char key_in[16];
	unsigned char key_out[16];
	uint16_t next_in;
	uint16_t next_out;

	int status;
	bool iscaller;
	bool initiated;
	bool secure;
	bool failed;
};



struct {
	object_t object;
	link_ctx *list[LINK_MAX_CONNECTIONS];
	int length;
} link_list;


void link_mixmaster(void);
void link_register(link_ctx *);
void link_unregister(link_ctx *);
void link_unregister_nm(link_ctx *);
void link_setsockopts(int);
void link_try_read(link_ctx *);
void link_try_write(link_ctx *);

link_ctx *link_create(int);
void link_free(void *);

bool link_do_read(link_ctx *,void *,size_t,bool);
bool link_do_write(link_ctx *,void *,size_t,bool);

packet_t *link_cache_get(link_ctx *);
void link_cache_put(link_ctx *,packet_t *);
packet_t *link_get(link_ctx *);
void link_put(link_ctx *,packet_t *);
int link_read(link_ctx *,unsigned char[PACKET_SIZE]);
int link_write(link_ctx *,unsigned char[PACKET_SIZE]);
int link_initiate(link_ctx *);
int link_respond(link_ctx *);
void link_close(link_ctx *);
bool link_is_closed(link_ctx *);

channel_ctx *channel_create(link_ctx *);
void channel_free(void *);

bool channel_do_read(channel_ctx *,void *,size_t,int);
bool channel_do_write(channel_ctx *,void *,size_t,int);

packet_t *channel_get(channel_ctx *);
void channel_put(channel_ctx *,packet_t *);
int channel_initiate(channel_ctx *);
int channel_respond(channel_ctx *);

#endif /* _LINK_H */
