/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <pthread.h>
#include <signal.h>

#include "common.h"
#include "options.h"
#include "log.h"
#include "thread.h"
#include "conf.h"
#include "crypt.h"
#include "pdetach.h"
#include "server.h"
#include "caller.h"
#include "node.h"
#include "packet.h"
#include "main.h"
#include "job.h"
#include "link.h"


int main(int argc, char *argv[]) {
	extern options_t options;
	extern conf_t conf;
	extern sig_atomic_t __conf_initialized;
	server_t *node_server;
	thread_t *sig_handler;
	thread_t *sched_thread;
	thread_t *mixmaster;
	sigset_t signals;

	atexit(exit_handler);

	options_init();
	options_parse(argc,argv);
		
	if (options.help) {
		usage();
		return EXIT_SUCCESS;
	} else if (options.license) {
		license();
		return EXIT_SUCCESS;
	} else if (options.version) {
		version();
		return EXIT_SUCCESS;
	}


	conf_init();
	conf_parse();


	if (!change_privileges()) {
		ERROR("there was a problem changing privileges.");
		return EXIT_FAILURE;
	}
		
	if (conf.daemon) {
		DEBUG("forking process.");
		if (!pdetach()) {
			ERROR("error forking.");
			return EXIT_FAILURE;
		}
	}



	log_init();
	rand_init();	/* init random data source */
	scheduler_init();
	ke_init();

	schedule((void (*)(void *))object_tracker,NULL,10,false,false);

	/* ignore all threads so our signal handler thread will catch them */
	sigfillset(&signals);
	pthread_sigmask(SIG_BLOCK,&signals,NULL);
	ASSERT((sig_handler = thread_create((void(*)(void *))signal_handler,NULL)) != NULL,"could not start signal thread");
	thread_run(sig_handler);

	ASSERT((sched_thread = thread_create((void(*)(void *))scheduler,NULL)) != NULL,"could not start scheduler thread");
	thread_run(sched_thread);

 
	INFO("anonnet started");

	ASSERT((mixmaster = thread_create((void(*)(void *))link_mixmaster,NULL)) != NULL,"could not start mixmaster thread");
	thread_run(mixmaster);


	if ((node_server = server_create(conf.port,link_setsockopts,node_call_back)) == NULL) {
		ERROR("could not get a node server");
		return EXIT_FAILURE;
	}


	thread_run(node_server);


	if (thread_is_alive(node_server) == false) {
		ERROR("could not start node server");
		return EXIT_FAILURE;
	}


	if (options.debug) {
		debug();
	}
	

	for (;;) {
		sleep(1);
		if (thread_is_alive(node_server) == false) {
			WARN("node server is not alive");
			return EXIT_FAILURE;
		}
	}

	return EXIT_SUCCESS;
}


void signal_handler(void) {
	sigset_t signals;
	int caught;

	sigfillset(&signals);

	DEBUG("started signal handler");

	for (;;) {

	//pthread_sigmask(SIG_UNBLOCK,&signals,NULL);
	sigwait(&signals,&caught);

	switch (caught) {
		case SIGHUP:		/* Hangup (POSIX).  */
			INFO("caught SIGHUP");
			conf_reload();
			break;
		case SIGINT:		/* Interrupt (ANSI).  */
			INFO("caught SIGINT");
			exit(EXIT_SUCCESS);
			break;
		case SIGQUIT:		/* Quit (POSIX).  */
			INFO("caught SIGQUIT");
			break;
		case SIGILL:		/* Illegal instruction (ANSI).  */
			INFO("caught SIGILL");
			break;
		case SIGTRAP:		/* Trace trap (POSIX).  */
			INFO("caught SIGTRAP");
			break;
		case SIGABRT:		/* Abort (ANSI).  */
			INFO("caught SIGABRT");
			break;
		case SIGKILL:		/* Kill, unblockable (POSIX).  */
			INFO("caught SIGKILL");
			break;
		case SIGUSR1:		/* User-defined signal 1 (POSIX).  */
			INFO("caught SIGUSR1");
			break;
		case SIGSEGV:		/* Segmentation violation (ANSI).  */
			INFO("caught SIGSEGV");
			break;
		case SIGUSR2:		/* User-defined signal 2 (POSIX).  */
			INFO("caught SIGUSR2");
			break;
		case SIGPIPE:		/* Broken pipe (POSIX). */
			INFO("caught SIGPIPE");
			break;
		case SIGALRM:		/* Alarm clock (POSIX). */
			INFO("caught SIGALRM");
			break;
		case SIGTERM:		/* Termination (ANSI).  */
			INFO("caught SIGTERM");
			break;
		case SIGSTOP:		/* Stop, unblockable (POSIX). */
			INFO("caught SIGSTOP");
			break;
		case SIGTSTP:		/* Keyboard stop (POSIX). */
			INFO("caught SIGTSTP");
			exit(EXIT_SUCCESS);
			break;
		default:
			INFO("ignoring unknown signal %i",caught);
	}		

	}
}


void exit_handler(void) {
	INFO("anonnet exiting");

	return;
}


/*
 * drop env variables? alter ulimit and scheduling priority?
 * login class?
 */
bool change_privileges() {
	struct group *group;
	struct passwd *passwd;
		
	/* set umask */
	umask(0);
	umask(027);
	

	if (conf.group != NULL) {
		if ((group = getgrnam(conf.group)) != NULL || (group = getgrgid((gid_t)atoi(conf.group))) != NULL) {
			if (setgid(group->gr_gid) != 0) {
				WARN("could not change to group %s: %s",group->gr_name,strerror(errno));
				return false;
			} else {
				INFO("changed to group %s",group->gr_name);
			}
		} else {
			WARN("could not change to group %s: invalid group or ENOMEM",conf.group);
			return false;
		}
	}
	
	if (conf.user != NULL) {
		if ((passwd = getpwnam(conf.user)) != NULL || (passwd = getpwuid((uid_t)atoi(conf.user))) != NULL) {
			if (setuid(passwd->pw_uid) != 0) {
				WARN("could not change to user %s: %s",passwd->pw_name,strerror(errno));
				return false;
			} else {
				INFO("changed to user %s",passwd->pw_name);
			}
		} else {
			WARN("could not change to user %s: invalid user or ENOMEM",conf.user);
			return false;
		}
	}
	
	return true;
}


void debug() {
	unsigned char request[1 + 256 + 2];
	caller_t *caller;
	int sv[2];

	socketpair(AF_LOCAL,SOCK_STREAM,0,sv);

	FLAG_CLR(request[0]);
	FLAG_SET(request[0],0x04); /* NAME */
	strcpy((char *)&request[1],"127.0.0.1");
	*((uint16_t *)&request[257]) = htons(80);

	if ((caller = caller_create(sv[1])) == NULL) {
		ERROR("could not get caller for debugging.");
		return;
	}

	thread_run(caller);

	write(sv[0],request,sizeof(request));


	return; /* void */
}



void usage() {
	printf("%s\n",USAGE);
}

void license() {
	printf("%s\n",LICENSE);
}

void version() {
	printf("%s %s\n",PACKAGE,VERSION);
}
